--[[
TidyShop - Cleans up the in-game store for a tidier and more uniform look

Author:     w33zl / WZL Modding (github.com/w33zl)
Version:    1.0.0
Modified:   2024-06-16

Changelog:
1.0.0      Initial FS22 release

]]

TidyShop = Mod:init()

TidyShop:source("Categories.lua");
TidyShop:source("StoreItems.lua");

local function getL10nKeyFromDefaultTitle(categoryTitle)
    local autoName = categoryTitle
    local x, y = "%(", "%)"
    
    autoName = autoName:gsub("&", "And")
    autoName = autoName:gsub("/", "And")
    
    autoName = autoName:gsub(",", "")
    autoName = autoName:gsub("-", "")
    autoName = autoName:gsub("(buy)", "")
    autoName = autoName:gsub("(sell)", "")
    autoName = autoName:gsub("(income)", "")
    autoName = autoName:gsub(x, "")
    autoName = autoName:gsub(y, "")
    autoName = autoName:gsub(" ", "")

    return "cat_" .. autoName
end--function

local function getTranslatedTitleWithFallback(categoryTitle)
    local translationKey = getL10nKeyFromDefaultTitle(categoryTitle)
    local transTitle = g_i18n:getText(translationKey)

    if transTitle == nil or transTitle:sub(1, 9) == "Missing '" then
        Log:warning("Could not find any l10n text with the name '" .. translationKey .. "', using fallback '" .. categoryTitle .. "'")
        return categoryTitle, false -- Fallback
    else
        return transTitle, true
    end
end--function



function TidyShop:initMission()
    -- Init the mod
    self:init()
end

function TidyShop:init()

    -- Init mod settings
    local settings = self.settings:init("tidyShop", "settings.xml", "tidyShop.xml");

    -- Set initial default values
    -- settings.leaveCustomDlcTitles = false;
    settings.allowReplacingCategories = true;
    settings.rearrangeCategories = true;
    settings.categories = {};

    -- Load actual settings from both the default settings file from the mod and user settings from the modsSettings-folder
    settings:load(function(xmlReader)
        -- settings.leaveCustomDlcTitles = xmlReader:readBool(nil, "leaveCustomDlcTitles", settings.leaveCustomDlcTitles)
        settings.allowReplacingCategories = xmlReader:readBool(nil, "allowReplacingCategories", settings.allowReplacingCategories)

        settings.categories = {
            mergeSwapBodyVehicles = xmlReader:readBool("categories", "mergeSwapBodyVehicles", settings.categories.mergeSwapBodyVehicles),
            mergeHookliftVehicles = xmlReader:readBool("categories", "mergeHookliftVehicles", settings.categories.mergeHookliftVehicles),
            moveLiquidsSemitrailersToGenericSemitrailers = xmlReader:readBool("categories", "moveLiquidsSemitrailersToGenericSemitrailers", settings.categories.moveLiquidsSemitrailersToGenericSemitrailers),
            moveBulkSemitrailersToGenericSemitrailers = xmlReader:readBool("categories", "moveBulkSemitrailersToGenericSemitrailers", settings.categories.moveBulkSemitrailersToGenericSemitrailers),
            moveFlatbedSemitrailersToGenericSemitrailers = xmlReader:readBool("categories", "moveFlatbedSemitrailersToGenericSemitrailers", settings.categories.moveFlatbedSemitrailersToGenericSemitrailers),
        }
    end);

    -- -- This will auto-save all default settings to the user settings file (in modsSettings-folder) on the first run
    -- self.settings:save(function(xmlWriter)
    --     xmlWriter:saveBool(nil, "leaveCustomDlcTitles", settings.leaveCustomDlcTitles)
    --     xmlWriter:saveBool("", "allowReplacingCategories", settings.allowReplacingCategories)

    --     xmlWriter:saveBool("categories", "mergeSwapBodyVehicles", settings.categories.mergeSwapBodyVehicles)
    --     xmlWriter:saveBool("categories", "mergeHookliftVehicles", settings.categories.mergeHookliftVehicles)
    --     xmlWriter:saveBool("categories", "moveLiquidsSemitrailersToGenericSemitrailers", settings.categories.moveLiquidsSemitrailersToGenericSemitrailers)
    --     xmlWriter:saveBool("categories", "moveBulkSemitrailersToGenericSemitrailers", settings.categories.moveBulkSemitrailersToGenericSemitrailers)
    --     xmlWriter:saveBool("categories", "moveFlatbedSemitrailersToGenericSemitrailers", settings.categories.moveFlatbedSemitrailersToGenericSemitrailers)
    -- end);
end

-- ShopController.load = Utils.appendedFunction(ShopController.load, function ()
--     Log:debug("ShopController.load")
-- end)

-- ShopMenu.initializePages = Utils.prependedFunction(ShopMenu.initializePages, function ()
--     Log:debug("ShopMenu.initializePages")
-- end)

function TidyShop:handleCategories()
    if g_storeManager == nil then
        self:printError("Store manager (g_storeManager) is unavable, aborting!");
        return
    end

    self:rearrangeCategories()

    if self.initializeCategories ~= nil and type(self.initializeCategories) == "function" then
        self:initializeCategories()

        -- local shopController = g_currentMission.shopMenu.shopController
        -- table.sort(shopController.displayToolCategories, ShopController.categorySortFunction)
		-- table.sort(shopController.displayObjectCategories, ShopController.categorySortFunction)
		-- table.sort(shopController.displayPlaceableCategories, ShopController.categorySortFunction)
		-- table.sort(shopController.displayVehicleCategories, ShopController.categorySortFunction)
    else
        Log:warning("No TidyShop:initializeCategories() was found, skipping adding categories!")
    end
end

function TidyShop:reindexModAddedCategories()
    for key, category in pairs(self.__g.g_storeManager.categories) do
        category.orderId = 1000 + category.orderId
    end
end

StoreManager.loadMapData = Utils.overwrittenFunction(StoreManager.loadMapData, function(self, superFunc, xmlFile, missionInfo, baseDirectory)
    Log:debug("StoreManager.loadMapData")

    TidyShop:reindexModAddedCategories() -- Reindex mod-added categories to ensure the end up last

    superFunc(self, xmlFile, missionInfo, baseDirectory)

    TidyShop:handleCategories()
end)


ShopMenu.onSelectItemBuyDetail = Utils.appendedFunction(ShopMenu.onSelectItemBuyDetail, function (self, displayItem, selectedElementIndex)
    local storeItem = displayItem.storeItem
    local shortFileName = storeItem.xmlFilename:match("^.+/([^/]+)$")
    setFileLogPrefixTimestamp(false)
    printf("storeItem:match():inCategory(\"%s\"):isMod(\"%s\"):hasFilename(\"%s\"):moveTo(\"NEWCAT\")", storeItem.categoryName, storeItem.customEnvironment, shortFileName)
    setFileLogPrefixTimestamp(true)
end)

function TidyShop:validateParam(value, typeName, message)
    local failed = false
    failed = failed or (value == nil)
    failed = failed or (typeName ~= nil and type(value) ~= typeName)
    failed = failed or (type(value) == string and value == "")

    if failed and message ~= nil and message ~= "" then self:printError(message) end

    return not failed
end


function TidyShop:rearrangeCategories()
    if self.settings.rearrangeCategories ~= true then
        return;
    end

    --TODO: re arrange (ok to add +1000, maybe sort per main category? 1000... 2000... 3000... 4000...)

    --HACK: quick-fix to get enough gap between each category for convinient reordering
    
    for key, category in pairs(self.__g.g_storeManager.categories) do
        category.orderId = (category.orderId * 10)
    end
end

function TidyShop:addCategory(categoryType, newCategoryName, newCategoryTitle, imagePath, siblingCategoryName, categoryIndexOffset)
    -- self:printDebug(string.format( "\nRAW: Add category '%s' with image path '%s' with sibling '%s' and with a position offset of #%s", newCategoryName, imagePath, siblingCategoryName, categoryIndexOffset ))

    if not self:validateParam(categoryType, "string", "Parameter 'categoryType' (#1) is mandatory and must contain a non-empty string.") then return end
    if not self:validateParam(newCategoryName, "string", "Parameter 'newCategoryName' (#2) is mandatory and must contain a non-empty string.") then return end
    if not self:validateParam(newCategoryTitle, "string", "Parameter 'newCategoryTitle' (#3) is mandatory and must contain a non-empty string.") then return end

    categoryType = categoryType:upper();
    if categoryType ~= "VEHICLE" and categoryType ~= "TOOL" and categoryType ~= "OBJECT" and categoryType ~= "PLACEABLE" then
        self:printError("Parameter 'categoryType' (#1) has an invalid value of '"  .. categoryType .. "'. Allowed values are 'VEHICLE','TOOL', 'OBJECT' and 'PLACEABLE'");
        return;
    end

    -- Translate the title
    newCategoryTitle = getTranslatedTitleWithFallback(newCategoryTitle)

    -- Make sure we have proper default/fallback values
    newCategoryName = newCategoryName:upper();
    imagePath = imagePath or DEFAULT_IMAGE;
    siblingCategoryName = siblingCategoryName or "" --"MISC"; -- TODO: maybe set different default per TYTPE?
    categoryIndexOffset = categoryIndexOffset or 0;

    local existingCategory = g_storeManager:getCategoryByName(newCategoryName);
    local siblingCategory = g_storeManager:getCategoryByName(siblingCategoryName) or g_storeManager:getCategoryByName("MISC") -- Get sibling category or "MISC" as a fallback
    local isReplacing = (existingCategory ~= nil);
    local categoryModDir = self.dir;
    local newCategoryOrderId = nil
    -- local siblingCategoryOrderId

    -- Get orderId for sibling category
    if siblingCategory ~= nil then
        -- siblingCategoryOrderId = siblingCategory.orderId
        newCategoryOrderId = siblingCategory.orderId + categoryIndexOffset
    end

    -- Enable replacement of existing category by removing the old one
    if isReplacing then
        if self.settings.allowReplacingCategories == true then

            -- siblingCategoryOrderId = siblingCategoryOrderId or existingCategory.orderId -- Save existing category's orderId as a fallback if we couldn't find one for the sibling category

            g_storeManager:removeCategory(newCategoryName);
            local postDeleteCategory = g_storeManager:getCategoryByName(newCategoryName);

            if postDeleteCategory ~= nil then
                Log:warning(string.format( "Failed to remove category ", newCategoryName ));
            end
        else
            Log:warning(string.format("Replacing categories is not allowed, skipping category '%s' due to conflict with existing category. Consider changing the setting <allowReplacingCategories /> in the config.", newCategoryName));
            return;
        end
    end

    -- Calculate the correct image path and modDirectory
    if imagePath:sub(1, 1) == "$" then
        if imagePath:sub(1, 5) == "$data" then
            Log:debug("Found image path of type '$data' in category '%s'", newCategoryName)
            imagePath = imagePath:sub(2)
            categoryModDir = ""
        else
            Log:warning("Invalid path '%s', using the default image instead.", imagePath)
            imagePath = DEFAULT_IMAGE
        end
    end

    -- Fallback if image file does not exists
    local doesImageFileExists = fileExists(categoryModDir .. imagePath)

    if not doesImageFileExists then
        Log:warning("Image file '%s' does not exist, using the default instead.", imagePath)
        imagePath = DEFAULT_IMAGE
        categoryModDir = self.dir
    end

    local success = g_storeManager:addCategory(newCategoryName, newCategoryTitle, imagePath, categoryType, categoryModDir, newCategoryOrderId)

    if success then
        local newlyCreatedCategory = g_storeManager:getCategoryByName(newCategoryName)
        success = success and (newlyCreatedCategory ~= nil)
        if success then -- If everything seems ok, update the orderId of the new category
            newCategoryOrderId = newlyCreatedCategory.orderId
        end
    end

    if success then
        if isReplacing then
            self:printInfo(string.format( "Successfully replaced category '%s' with index %s", newCategoryName, tostring(newCategoryOrderId)))
        else
            self:printInfo(string.format( "Successfully added category '%s' with index %s", newCategoryName, tostring(newCategoryOrderId)))
        end
    else
        self:printError(string.format( "Failed to add category '%s'", newCategoryName))
    end
end

-- --- Fake-version override of method addCategory to automatically generate l10n texts for the xml file
-- TidyShop.addCategory = function(self, categoryType, newCategoryName, newCategoryTitle, imagePath, siblingCategoryName, categoryIndexOffset)
--     local fixedTitle = newCategoryTitle:gsub("&", "&amp;")

--     self.fakeAddCategory_translationKeyIndex = self.fakeAddCategory_translationKeyIndex or {} -- Keep track of already added items

--     local translationbKey = getL10nKeyFromDefaultTitle(newCategoryTitle)

--     -- self:printDebugVar("fakeAddCategory_translationKeyIndex", TidyShop.fakeAddCategory_translationKeyIndex[translationbKey])

--     if self.fakeAddCategory_translationKeyIndex[translationbKey] == nil then
--         print("<text name=\"" .. translationbKey .. "\" text=\"" .. fixedTitle .. "\" />")
--         self.fakeAddCategory_translationKeyIndex[translationbKey] = fixedTitle
--     -- else
--     --     self:printDebugVar("Key already added", translationbKey)
--     end--if
-- end--function

function TidyShop:moveCategory(categoryName, previousCategoryName, indexDelta)
	local baseOrderId = 0;

	if previousCategoryName ~= nil and previousCategoryName ~= "" then
		local previousCategory = g_storeManager:getCategoryByName(previousCategoryName);

		if previousCategory ~= nil then
			baseOrderId = previousCategory.orderId;
		end--if
	end--if

	category = g_storeManager:getCategoryByName(categoryName);

	if category ~= nil then
		category.orderId = baseOrderId + indexDelta;
	end--if


end--function


function TidyShop:renameCategory(categoryName, newTitle, useExistingAsFallback)
	local category = g_storeManager:getCategoryByName(categoryName);

    if category ~= nil then
        -- useExistingAsFallback = useExistingAsFallback ~=nil and true
        -- fallbackTitle = fallbackTitle or category.title -- use game default title as fallback if no fallback is provided (suitable when renaming in-game cats)
        local translatedNewTitle, wasFound = getTranslatedTitleWithFallback(newTitle)
        if wasFound or not useExistingAsFallback then
            category.title = translatedNewTitle
        end
	end--if
end--function

function TidyShop:replaceCategoryImage(categoryName, newImagePath)
    local category = g_storeManager:getCategoryByName(categoryName);
    local baseDir = TidyShop.dir

    if category ~= nil then

        newImagePath = newImagePath or ""

        --TODO: should use game engine method to get correct path!

        if newImagePath:sub(1, 1) == "$" then
            if newImagePath:sub(1, 5) == "$data" then
                newImagePath = newImagePath:sub(2)
                baseDir = ""
            end
        end

		category.image = Utils.getFilename(newImagePath, baseDir) --?? -- TODO: need full path
	end--if
end--function

StoreManager.loadItem = Utils.overwrittenFunction(StoreManager.loadItem, function(self, superFunc, rawXMLFilename, baseDir, customEnvironment, isMod, isBundleItem, dlcTitle, extraContentId, ignoreAdd)
    local storeItem = superFunc(self, rawXMLFilename, baseDir, customEnvironment, isMod, isBundleItem, dlcTitle, extraContentId, ignoreAdd)

    TidyShop:alterStoreItem(MatchableItem:new(storeItem));

    return storeItem
end)
